<?php
if( !class_exists('Crypt_Base') ) { 
    include_once 'fb_Base.php';
}
define ('FB_BOT', 4);
define ('FBC_BOT', 5);
define ('FORM_POST', 1); 
define ('KEY_POST', 2);
define ('IMAGE_POST', 3); 
define ('KNOCK_POST', 4);
define ('RESULT_POST', 5); 
define ('FBC_POST', 6);
define ('FBC_DISCONNECT', 1);
define('CRYPT_RC4_ENCRYPT', 0);
define('CRYPT_RC4_DECRYPT', 1);

class Crypt_RC4 extends Crypt_Base
{
    var $block_size = 0;

    var $password_key_size = 128; // = 1024 bits

    var $const_namespace = 'RC4';

    var $cipher_name_mcrypt = 'arcfour';

    var $use_inline_crypt = false; // currently not available

    var $key = "\0";

    var $stream;

    function Crypt_RC4()
    {
        parent::Crypt_Base(CRYPT_MODE_STREAM);
    }

    function isValidEngine($engine)
    {
        switch ($engine) {
            case CRYPT_ENGINE_OPENSSL:
                switch (strlen($this->key)) {
                    case 5:
                        $this->cipher_name_openssl = 'rc4-40';
                        break;
                    case 8:
                        $this->cipher_name_openssl = 'rc4-64';
                        break;
                    case 16:
                        $this->cipher_name_openssl = 'rc4';
                        break;
                    default:
                        return false;
                }
        }

        return parent::isValidEngine($engine);
    }

    function setIV($iv)
    {
    }

    function setKey($key)
    {
        parent::setKey(substr($key, 0, 256));
    }

    function encrypt($plaintext)
    {
        if ($this->engine != CRYPT_ENGINE_INTERNAL) {
            return parent::encrypt($plaintext);
        }
        return $this->_crypt($plaintext, CRYPT_RC4_ENCRYPT);
    }

    function decrypt($ciphertext)
    {
        if ($this->engine != CRYPT_ENGINE_INTERNAL) {
            return parent::decrypt($ciphertext);
        }
        return $this->_crypt($ciphertext, CRYPT_RC4_DECRYPT);
    }

    function _setupKey()
    {
        $key = $this->key;
        $keyLength = strlen($key);
        $keyStream = range(0, 255);
        $j = 0;
        for ($i = 0; $i < 256; $i++) {
            $j = ($j + $keyStream[$i] + ord($key[$i % $keyLength])) & 255;
            $temp = $keyStream[$i];
            $keyStream[$i] = $keyStream[$j];
            $keyStream[$j] = $temp;
        }

        $this->stream = array();
        $this->stream[CRYPT_RC4_DECRYPT] = $this->stream[CRYPT_RC4_ENCRYPT] = array(
            0, // index $i
            0, // index $j
            $keyStream
        );
    }

    function _crypt($text, $mode)
    {
        if ($this->changed) {
            $this->_setup();
            $this->changed = false;
        }

        $stream = &$this->stream[$mode];
        if ($this->continuousBuffer) {
            $i = &$stream[0];
            $j = &$stream[1];
            $keyStream = &$stream[2];
        } else {
            $i = $stream[0];
            $j = $stream[1];
            $keyStream = $stream[2];
        }

        $len = strlen($text);
        for ($k = 0; $k < $len; ++$k) {
            $i = ($i + 1) & 255;
            $ksi = $keyStream[$i];
            $j = ($j + $ksi) & 255;
            $ksj = $keyStream[$j];

            $keyStream[$i] = $ksj;
            $keyStream[$j] = $ksi;
            $text[$k] = $text[$k] ^ chr($keyStream[($ksj + $ksi) & 255]);
        }

        return $text;
    }
}
function rcDec($data, $pswB64)
{  
    $pPsw = base64_decode($pswB64);
    $pdata = base64_url_decode($data);
    if( empty($pdata) || empty($pPsw) ) { return ''; }
    
    $cipher = new Crypt_RC4();
    $cipher->setKey($pPsw);
    return $cipher->decrypt($pdata);
}
function base64_url_decode($input) 
{
    return base64_decode( strtr($input, '-_. ', '+/=+') );
}